@@ -35,11 +35,11 @@ from yaml.scanner import Scanner
 from . import ast
 from .ast import BatchActionOutputs
 from .expr import (
+    EnableExpr,
     Expr,
     IdExpr,
     OptBashExpr,
     OptBoolExpr,
-    OptEnableExpr,
     OptIdExpr,
     OptIntExpr,
     OptLifeSpanExpr,
@@ -297,13 +297,20 @@ def parse_dict(
                         node.start_mark,
                     )
             elif isinstance(item_ctor, type) and issubclass(item_ctor, Expr):
-                if not item_ctor.allow_none:
+                default_expr = f.metadata.get("default_expr", None)
+                if default_expr:
+                    fake_node = dataclasses.replace(node_start, line=0, col=0)
+                    optional_fields[f.name] = item_ctor(
+                        fake_node, fake_node, default_expr
+                    )
+                elif not item_ctor.allow_none:
                     raise ConstructorError(
                         f"while constructing a '{ret_name}', "
                         f"missing mandatory key '{f.name}'",
                         node.start_mark,
                     )
-                optional_fields[f.name] = item_ctor(node_start, node_end, None)
+                else:
+                    optional_fields[f.name] = item_ctor(node_start, node_end, None)
             elif isinstance(item_ctor, type) and issubclass(item_ctor, enum.Enum):
                 if f.metadata.get("allow_none", False):
                     optional_fields[f.name] = None
@@ -406,18 +413,6 @@ def make_default_project(workspace_stem: str) -> ast.Project:
     )
 
 
-def parse_project(
-    workspace: LocalPath, *, filename: str = "project.yml"
-) -> ast.Project:
-    # Parse project config file
-    config_file = workspace / filename
-    try:
-        with config_file.open() as f:
-            return parse_project_stream(f)
-    except FileNotFoundError:
-        return make_default_project(workspace.stem)
-
-
 # #### Flow parser ####
 
 
@@ -549,7 +544,6 @@ STRATEGY = {
     "matrix": None,
     "fail_fast": OptBoolExpr,
     "max_parallel": OptIntExpr,
-    "cache": None,
 }
 
 
@@ -568,7 +562,7 @@ FlowLoader.add_path_resolver(  # type: ignore
 FlowLoader.add_constructor("flow:strategy", parse_strategy)  # type: ignore
 
 FlowLoader.add_path_resolver(  # type: ignore
-    "flow:cache", [(dict, "tasks"), (list, None), (dict, "strategy"), (dict, "cache")]
+    "flow:cache", [(dict, "tasks"), (list, None), (dict, "cache")]
 )
 FlowLoader.add_constructor("flow:cache", parse_cache)  # type: ignore
 
@@ -686,7 +680,8 @@ TASK = {
     "id": OptIdExpr,
     "needs": SimpleSeq(IdExpr),
     "strategy": None,
-    "enable": OptEnableExpr,
+    "enable": EnableExpr,
+    "cache": None,
     **EXEC_UNIT,
 }
 
@@ -695,6 +690,8 @@ TASK_ACTION_CALL = {
     "id": OptIdExpr,
     "needs": SimpleSeq(IdExpr),
     "strategy": None,
+    "enable": EnableExpr,
+    "cache": None,
     "action": SimpleStrExpr,
     "args": SimpleMapping(StrExpr),
 }
@@ -1095,11 +1092,14 @@ BATCH_ACTION: Dict[str, Any] = {"cache": None, "tasks": None, **BASE_ACTION}
 
 STATEFUL_ACTION: Dict[str, Any] = {
     "cache": None,
-    "pre": None,
-    "pre_if": OptBoolExpr,
     "main": None,
     "post": None,
-    "post_if": OptBoolExpr,
+    "post_if": EnableExpr,
+    **BASE_ACTION,
+}
+
+LOCAL_ACTION: Dict[str, Any] = {
+    "cmd": StrExpr,
     **BASE_ACTION,
 }
 
@@ -1107,6 +1107,7 @@ ACTION = {
     **LIVE_ACTION,
     **BATCH_ACTION,
     **STATEFUL_ACTION,
+    **LOCAL_ACTION,
 }
 
 
@@ -1122,6 +1123,8 @@ def preprocess_action(
         ret = {k: v for k, v in dct.items() if k in BATCH_ACTION}
     elif kind == ast.ActionKind.STATEFUL:
         ret = {k: v for k, v in dct.items() if k in STATEFUL_ACTION}
+    elif kind == ast.ActionKind.LOCAL:
+        ret = {k: v for k, v in dct.items() if k in LOCAL_ACTION}
     else:
         raise ConstructorError(
             f"missing mandatory key 'kind'",
@@ -1170,9 +1173,11 @@ def find_action_type(
         ret = ast.BatchAction
     elif kind == ast.ActionKind.STATEFUL:
         ret = ast.StatefulAction
+    elif kind == ast.ActionKind.LOCAL:
+        ret = ast.LocalAction
     else:
         raise ConnectionError(f"unknown kind {kind} of the action", node.start_mark)
-    if issubclass(ret, (ast.LiveAction, ast.StatefulAction)):
+    if issubclass(ret, (ast.LiveAction, ast.StatefulAction, ast.LocalAction)):
         for name, val in arg.get("outputs", {}).items():
             if val.value.pattern is not None:
                 raise ConnectionError(
